var settings = {};
var flags = {};

var holdOptionsTimer = null;

var reloadOnErrorTimer = null;
var reloadOnErrorTimeout = 8;
var openOptionRetries = 4;
var firstLoad = true;
var loadingPageOn = false;
var resumeReload = false;
var errorCount = 0;
var prevHistLen = 0;
var backCount = 0;

var unixUser = null;
var unixPass = null;

var bookmarkFolderName = "extBookmarksFoder";
var bookmarkFolderId = null;
var bookmarkTree = null;

var lastUrl = ""

function extLog(msg) {
   console.log("EXT_LOG: " + msg);
}

// Clear cache
chrome.browsingData.removeCache({});

// Parse command line flags
var cmdFlags = cmd_line.split(" ");
var defaultConfig = browser_settings;
var cmd;
var value;

for (i = 0; i < cmdFlags.length; i++) {

   if ( ! cmdFlags[cmdFlags.length-1].startsWith("--") )
      flags.page = cmdFlags[cmdFlags.length-1];

   cmd = cmdFlags[i].split('=', 1)[0];
   value = cmdFlags[i].split('=').slice(1).join('=');

   switch (cmd) {
      case "--disable-login":
         flags.disableLogin = true;
         break;
      case "--enable-exit-button":
         flags.enableExitButton = "normal";
         break;
      case "--enable-window-exit-button":
         flags.enableExitButton = "window";
         break;
      case "--enable-restart-button":
         flags.enableExitButton = "restart";
         break;
      case "--enable-toolbar":
         flags.enableToolbar = true;
         break;
      case "--enable-toolbar-on-error":
         flags.showToolbarOnError = true;
         break;
      case "--disable-toolbar":
         flags.enableToolbar = false;
         break;
      case "--enable-toolbar-bookmarks":
         flags.toolbarBookmarks = true;
         break;
      case "--disable-toolbar-reload":
         flags.toolbarReload = false;
         break;
      case "--disable-toolbar-options":
         flags.toolbarOptions = false;
         break;
      case "--disable-options":
         flags.toolbarOptions = false;
         flags.disableOptions = true;
         break;
      case "--disable-toolbar-url":
         flags.toolbarUrl = false;
         break;
      // Currently broken and undocumented
      case "--fixed-toolbar":
         flags.fixedToolbar = true;
         break;
      case "--open-options-after-num-retries":
         if ( value && ! isNaN(value) )
            openOptionRetries = value;
         break;
      case "--retries-on-error-timeout":
         if ( value && ! isNaN(value) )
            reloadOnErrorTimeout = value < 5 ? 5 : value;
         break;
      case "--loading-page":
         if ( value )
            flags.useLoadingPage = value;
         else
            flags.useLoadingPage = "";
         break;
      case "--fallback-url":
         if ( value )
            flags.fallbackPage = value;
         break;
      case "--homepage":
         if ( value )
            flags.homepage = value;
         break;
      case "--first-load-retry":
         flags.firstLoadRetry = true;
         break;
      case "--disable-fallback-reload":
         flags.disableFallbackReload = true;
         break;
   }
}

// Load settings from chromium local storage
chrome.storage.local.get(null, function(obj) {
   settings = obj;

   if ( flags.homepage ) {
      settings.homepage = flags.homepage;
      settings.onStartup = "homepage";
      chrome.storage.local.set( { homepage: settings.homepage, onStartup: settings.onStartup } );
   } else if( defaultConfig.homepageLink ){
       settings.homepage = defaultConfig.homepageLink;
       chrome.storage.local.set( { homepage: settings.homepage } );
   }

   if ( !settings.homepage ) {
        settings.homepage = "https://www.google.com";
        chrome.storage.local.set( { homepage: settings.homepage } );
   }

   if ( typeof flags.enableToolbar == "boolean" ) {
       settings.enableNavBar = flags.enableToolbar;
       chrome.storage.local.set( { enableNavBar: settings.enableNavBar } );
   } else if ( typeof defaultConfig.enableToolbar == "boolean" ) {
       settings.enableNavBar = defaultConfig.enableToolbar;
       chrome.storage.local.set( { enableNavBar: settings.enableNavBar } );
   }

   if ( typeof flags.disableFallbackReload  == "boolean" ) {
       settings.disableFallbackReload = flags.disableFallbackReload;
       chrome.storage.local.set( { disableFallbackReload: settings.disableFallbackReload } );
   } else if ( typeof defaultConfig.disableFallbackReload == "boolean" ) {
       settings.disableFallbackReload = defaultConfig.disableFallbackReload;
       chrome.storage.local.set( { disableFallbackReload: settings.disableFallbackReload } );
   }

   if ( typeof flags.useLoadingPage == "string" ) {
      settings.useLoadingPage = flags.useLoadingPage;
      chrome.storage.local.set( { useLoadingPage: settings.useLoadingPage } );
   } else if ( typeof defaultConfig.useLoadingPage == "string" ) {
      settings.useLoadingPage = defaultConfig.useLoadingPage;
      chrome.storage.local.set( { useLoadingPage: settings.useLoadingPage } );
   }

   if (settings.useLoadingPage == "") {
      settings.useLoadingPage = "file://" + inst_dir + "/ext/loading.html";
      chrome.storage.local.set( { useLoadingPage: settings.useLoadingPage } );
   }

   if ( !settings.lastVisitedPage ) {
      settings.lastVisitedPage = settings.homepage;
      chrome.storage.local.set( { lastVisitedPage: settings.homepage } );
   }

   if( defaultConfig.customUserAgent ){
      settings.customUserAgent = defaultConfig.customUserAgent;
      chrome.storage.local.set( { customUserAgent: settings.customUserAgent } );
   }
   if ( !settings.customUserAgent ) {
      settings.customUserAgent = null;
      chrome.storage.local.set( { customUserAgent: settings.customUserAgent } );
   }

   if ( typeof flags.showToolbarOnError  == "boolean" ) {
       settings.autoNavBar = flags.showToolbarOnError;
       chrome.storage.local.set( { autoNavBar: settings.autoNavBar } );
   } else if ( typeof defaultConfig.showToolbarOnError == "boolean" ) {
       settings.autoNavBar = defaultConfig.showToolbarOnError;
       chrome.storage.local.set( { autoNavBar: settings.autoNavBar } );
   }

   if ( !settings.autoNavBar ) {
      settings.autoNavBar = false;
      chrome.storage.local.set( { autoNavBar: settings.autoNavBar } );
   }

   if( typeof defaultConfig.showWiFiStatus == "boolean" ){
      settings.wifiStatus = defaultConfig.showWiFiStatus;
      chrome.storage.local.set( { wifiStatus: settings.wifiStatus } );
  }
   if ( !settings.wifiStatus ) {
      settings.wifiStatus = false;
      chrome.storage.local.set( { wifiStatus: settings.wifiStatus } );
   }

   if( typeof defaultConfig.showHistoryButton == "boolean" ){
      settings.historyBtns = defaultConfig.showHistoryButton;
      chrome.storage.local.set( { historyBtns: settings.historyBtns } );
   }
   if ( !settings.historyBtns ) {
      settings.historyBtns = false;
      chrome.storage.local.set( { historyBtns: settings.historyBtns } );
   }

   if( typeof defaultConfig.showLoadingBarAndStopButton == "boolean" ){
      settings.loadingBar = defaultConfig.showLoadingBarAndStopButton;
      chrome.storage.local.set( { loadingBar: settings.loadingBar } );
   }
   if ( !settings.loadingBar ) {
      settings.loadingBar = false;
      chrome.storage.local.set( { loadingBar: settings.loadingBar } );
   }

   if( defaultConfig.holdTimer ){
      if (defaultConfig.holdTimer < 2000)
         settings.holdOptionsTimerMilliseconds = 2000;
      else
         settings.holdOptionsTimerMilliseconds = defaultConfig.holdTimer;
      chrome.storage.local.set( { holdOptionsTimerMilliseconds: settings.holdOptionsTimerMilliseconds } );
   }

   if ( defaultConfig.onStartup == "settings" || defaultConfig.onStartup == "lastVisitedPage" || defaultConfig.onStartup == "homepage" ) {
      settings.onStartup = defaultConfig.onStartup;
      chrome.storage.local.set( { onStartup: settings.onStartup } );
   }

   if ( !settings.holdOptionsTimerMilliseconds ) {
      settings.holdOptionsTimerMilliseconds = 3500;
      chrome.storage.local.set( { holdOptionsTimerMilliseconds: settings.holdOptionsTimerMilliseconds } );
   }

   if (flags.fallbackPage) {
       settings.fallbackPage = flags.fallbackPage;
       chrome.storage.local.set( { fallbackPage: settings.fallbackPage } );
   } else if( defaultConfig.fallbackToDefaultPageLink ){
       settings.fallbackPage = defaultConfig.fallbackToDefaultPageLink;
       chrome.storage.local.set( { fallbackPage: settings.fallbackPage } );
   }
   if ( !settings.fallbackPage) {
       settings.fallbackPage = null;
       chrome.storage.local.set( { fallbackPage: settings.fallbackPage } );
   }

   lastUrl = settings.lastVisitedPage;
});

/*
 * This is called every time a tab is created. We want to keep the number of tabs
 * under control. We have a "navigation" tab ( index 0 ) and a "settings" tab ( index 1 )
 */
chrome.tabs.onCreated.addListener(function(tab) {

   if ( tab.id && tab.index == 0 && tab.url == "chrome://newtab/" ) {
      if( !settings.onStartup || settings.onStartup == "settings") {
         chrome.tabs.update( tab.id, { url: "options.html" });
      } else {
         var firstUrl;
         if ( settings.onStartup == "homepage" || flags.homepage )
            firstUrl = settings.homepage;
         else // lastVisitedPage
            firstUrl = settings.lastVisitedPage;

         if ( settings.useLoadingPage && firstLoad == true ) {
            loadingPageOn = true;
            chrome.tabs.update( tab.id, { url: settings.useLoadingPage });
            chrome.tabs.create( { url: firstUrl, active: false } );
         } else {
            chrome.tabs.update( tab.id, { url: firstUrl });
         }
      }
      return;
   }

   if ( ! tab.openerTabId )
      return;

   if ( tab.active ) { 

      // Replace the old with the new one
      chrome.tabs.remove(tab.openerTabId);
      chrome.tabs.move(tab.id, {index: 0});

      // Close also the settings tab if open
      chrome.tabs.query( { index: 1 }, function (tabs) {
         if ( tabs[0] )
            chrome.tabs.remove(tabs[0].id);
      });
   }
   else {
      // If it's a secondary tab, remove it
      chrome.tabs.remove(tab.id);
   }
});

/*
 * If login page is disabled we use Jmuconfig's HTTP basic login instead.
 * Here we handle login fails ( code 401 unauthorized )
 */
if ( flags.disableLogin ) {
   chrome.webRequest.onResponseStarted.addListener(function (obj) {

      if (obj && obj.tabId == 1 && obj.statusCode == 401)
         openOptions();

   }, {urls: ["http://127.0.0.1/*"]});
}

/*
 * Load bookmarks
 */
chrome.bookmarks.search({ title: bookmarkFolderName }, function(searchResults) {

   if ( !bookmarkTree == null )
      return;

   if (searchResults[0]) {
      bookmarkFolderId = searchResults[0].id;
      chrome.bookmarks.getSubTree(bookmarkFolderId, function(bookmarks){
         bookmarkTree = bookmarks[0].children;
      });
   } else {
      chrome.bookmarks.create({ title: bookmarkFolderName }, function (createResult) {
         if (createResult)
            bookmarkFolderId = createResult.id;
      });
   }

   // Notify content script about bookmarks
   chrome.bookmarks.onCreated.addListener(notifyBookmarkListChanged);
   chrome.bookmarks.onRemoved.addListener(notifyBookmarkListChanged);
});

function notifyBookmarkListChanged() {
   chrome.bookmarks.getSubTree(bookmarkFolderId, function(bookmarks){
      bookmarkTree = bookmarks[0].children;
      chrome.tabs.query( { index: 0 }, function (tabs) {
         if ( tabs[0] )
            chrome.tabs.sendMessage(tabs[0].id, { cmd: "updateBookmarks", tree: bookmarkTree });
      });
   });
}

function handleResumeReload() {
   if(resumeReload) {
      resumeReload = false;
      if (!reloadOnErrorTimer) {
         if ( loadingPageOn ) {
            chrome.tabs.query( { index: 1 }, function (tabs) {
               if (tabs[0])
                  reloadOnErrorTimer = window.setInterval(onErrorTimedOut(tabs[0]), 500);
            });
         } else {
            chrome.tabs.query( { active: true }, function (tabs) {
               if (tabs[0])
                  reloadOnErrorTimer = window.setInterval(onErrorTimedOut(tabs[0]), 500);
            });
         }
      }
   }
}

/*
 * Receive messages from content scripts
 */
chrome.runtime.onMessage.addListener(function(msg, sender, sendResponse) {

   switch (msg.cmd) {

      case "back":

         if ( msg.url && reloadOnErrorTimer) {
            window.clearTimeout(reloadOnErrorTimer);
            reloadOnErrorTimer = null;
         }

         errorCount = 0;

         chrome.tabs.query( { index: 0 }, function (tabs) {
            if ( tabs[0] ) {
               if ( msg.url )
                  chrome.tabs.update( tabs[0].id, { url: msg.url });
               else
                  chrome.tabs.sendMessage(tabs[0].id, { cmd: "updateSettings", settings: settings });
            }
         });

         chrome.tabs.query( { index: 1 }, function (tabs) {
            if ( tabs[0] )
               chrome.tabs.remove(tabs[0].id);
         });

         unixUser = null;
         unixPass = null;
         break;

      case "holdTimerStart":
         if (!holdOptionsTimer)
            holdOptionsTimer = window.setTimeout( function() {
               openOptions();
            }, settings.holdOptionsTimerMilliseconds);
         break;

      case "holdTimerStop":
         // The timeout is necessary here because messanges seems to come
         // out of order sometimes, meaning that we could stop the timer
         // before starting it.
         window.setTimeout(function(){
            if(holdOptionsTimer)
               window.clearTimeout(holdOptionsTimer);
            holdOptionsTimer = null;
         }, 200);
         break;

      case "loadPage":
         errorCount = 0;

         chrome.tabs.query( { index: 1 }, function (tabs) {
            if ( tabs[0] )
               chrome.tabs.remove(tabs[0].id);
            loadingPageOn = false;
         });
         if(msg.url)
              chrome.tabs.update( sender.tab.id, { url: msg.url });
         else
              chrome.tabs.reload( sender.tab.id);
         break;

      case "openOptions":
         openOptions();
         break;

      case "login":
         if (msg.user && msg.pass) {
            unixUser = msg.user;
            unixPass = msg.pass;
         }
         openOnSecondTab("chrome://settings");
         break;

      case "jmuconfig":
         openJMuconfig();
         break;

      case "getBookmarkTree":
         sendResponse(bookmarkTree);
         break;

      case "addBookmark":
         chrome.tabs.query( { index: 0 }, function (tabs) {
            if ( tabs[0] )
               chrome.bookmarks.create({ title: msg.title, url: tabs[0].url, parentId: bookmarkFolderId });
         });
         break;

      case "removeBookmark":
         chrome.bookmarks.remove( msg.id );
         break;

      case "setSettings":
         Object.assign(settings, msg.settings);
         chrome.storage.local.clear(function() {
            chrome.storage.local.set(settings);
         });
         break;

      case "getInfo":
            sendResponse({ loadingPage: loadingPageOn, tabIndex: sender.tab.index, settings: settings, flags: flags });
         break;

      case "clearCache":
         clearCacheAndReload();
         break;

      case "closeNavBar":
         chrome.tabs.query( { index: 0 }, function (tabs) {
            if ( tabs[0] )
               chrome.tabs.sendMessage(tabs[0].id, { cmd: "updateSettings", closeNavBar: true });
         });
         break;

      case "exitChrome":
         if ( flags.enableExitButton == "window" )
            chrome.windows.getAll(null, function(wins) {
               wins.forEach(function(win) {
                  chrome.windows.remove(win.id);
               });
            });
         else if ( flags.enableExitButton == "restart" ) {
            var xmlHttp = new XMLHttpRequest();
            xmlHttp.open("POST", "http://127.0.0.1:3000/api/v1/system/jmlauncher", true);
            xmlHttp.setRequestHeader("Content-Type", "application/json;charset=utf-8");
            xmlHttp.send('{ "action": "restart", "appid": "Chromium" }');
         }
         else if ( flags.enableExitButton )
            chrome.processes.terminate(0);
         break;

      case "pauseReload":
         if (reloadOnErrorTimer) {
            window.clearTimeout(reloadOnErrorTimer);
            reloadOnErrorTimer = null;
            resumeReload = true;
         }
         break;

      case "resumeReload":
         handleResumeReload();
         break;

      case "getBackCount":
         sendResponse({ "backCount": backCount });
         break;

      case "goingBack":
         backCount += 2;
         break;

      case "openingPage":
         backCount -= 1;
         if (backCount < 0)
            backCount = 0;
         if (prevHistLen < msg.hlen)
            backCount = 0;
         prevHistLen = msg.hlen;
         break;

      default:
         return false;
   }

   return true;
});

/*
 * This is called upon network errors. Here we manage the autoreload on error
 */
chrome.webNavigation.onErrorOccurred.addListener(function(details) {

   if( details.frameId != 0 || openOptionRetries < 0  || errorCount < 0 )
      return;

   if ( details.error == "net::ERR_NAME_NOT_RESOLVED" || details.error == "net::ERR_ADDRESS_UNREACHABLE"
         || details.error == "net::ERR_INTERNET_DISCONNECTED" || details.error == "net::ERR_NETWORK_CHANGED"
         || details.error == "net::ERR_CERT_INVALID" || details.error == "net::ERR_CONNECTION_REFUSED" || details.error == "net::ERR_FILE_NOT_FOUND" || details.error == "net::ERR_INTERNET_DISCONNECTED" ) 
{

      chrome.tabs.get(details.tabId, function(tab) {

         if ( !tab || ( loadingPageOn && tab.index != 1 ) || ( !loadingPageOn && tab.index != 0 ))
            return;

         // On certificate error the user is asked to confirm he wants to proceed. Stop reloading the page
         if ( details.error == "net::ERR_CERT_INVALID" ) {
            stopReloadOnError();
            return;
         }

         if (!reloadOnErrorTimer && !resumeReload) {
            reloadOnErrorTimer = window.setInterval(onErrorTimedOut, reloadOnErrorTimeout * 1000, details);
         }
      });
   }
});

function onErrorTimedOut(details) {

   if ( loadingPageOn ) {
      chrome.tabs.query( { index: 0 }, function (tabs) {
         chrome.tabs.sendMessage(tabs[0].id, { cmd: "loadError", error: details.error });
      });
   }

   if ( errorCount == (openOptionRetries -1) && errorCount > 0 ) {
      if ( settings.fallbackPage /*&& firstLoad*/ ) {
         window.clearTimeout(reloadOnErrorTimer);
         reloadOnErrorTimer = null;
         firstLoad = false;
         resumeReload = false;

         if (settings.disableFallbackReload)
            errorCount = -1;

         if ( loadingPageOn ) {
            chrome.tabs.query( { index: 1 }, function (tabs) {
               if (tabs[0])
                  chrome.tabs.remove(tabs[0].id);
            });

            chrome.tabs.query( { index: 0 }, function (tabs) {
               if (tabs[0])
                  chrome.tabs.update(tabs[0].id, { url: settings.fallbackPage });
            });
            loadingPageOn = false;
         } else {
            chrome.tabs.update(details.tabId, { url: settings.fallbackPage });
         }

         return;

      } else if (!flags.disableOptions)
         openOptions();
   }

   chrome.tabs.query( { index: loadingPageOn ? 1 : 0 }, function (tabs) {
      chrome.tabs.reload(tabs[0].id);
   });

   if ( openOptionRetries == 0 || errorCount >= openOptionRetries ||  errorCount < 0 )
      return;

   errorCount = errorCount + 1;
}

/*
 * Stop the autoreload as soon as we successfully load a page and save last visited url
 */
chrome.webNavigation.onCommitted.addListener(function(details) {

   chrome.tabs.get(details.tabId, function(tab) {
      if (!tab)
          return;

      if (( loadingPageOn && tab.index == 1 ) || ( !loadingPageOn && tab.index == 0 ))
          stopReloadOnError();

      if( tab.index != 0 || details.url.startsWith("chrome") || loadingPageOn || tab.url == lastUrl )
          return;

      lastUrl = tab.url;
      chrome.storage.local.set({ lastVisitedPage: tab.url });
   });
});

function stopReloadOnError() {

    firstLoad = false;
    errorCount = 0;
    resumeReload = false;

    if ( flags.firstLoadRetry )
        openOptionRetries = -1;

    if ( reloadOnErrorTimer ) {
        window.clearTimeout(reloadOnErrorTimer);
        reloadOnErrorTimer = null;
    }

    if ( loadingPageOn ) {
        chrome.tabs.query( { index: 0 }, function (tabs) {
           if ( tabs[0] )
              window.setTimeout(function() {
                 chrome.tabs.remove(tabs[0].id);
              }, 5000);
           loadingPageOn = false;
        });
    }
}

/*chrome.tabs.onRemoved.addListener(function(tabId, removeInfo) {
   errorCount = 0;
});
*/

/*
 * Open the settings tab on given url
 */
function openOnSecondTab(url) {

    if ( loadingPageOn ) {
        loadingPageOn = false;
        chrome.tabs.query( { index: 0 }, function (tabs) {
            if ( tabs[0] )
               chrome.tabs.remove(tabs[0].id);

            chrome.tabs.query( { index: 1 }, function (tabs) {
                if ( tabs[0] ) {
                    if ( tabs[0].active ) {
                        chrome.tabs.update( tabs[0].id, { url: url });
                        return;
                    } else {
                        chrome.tabs.remove(tabs[0].id);
                    }
                }
                if ( firstLoad )
                    errorCount = openOptionRetries +1;

                chrome.tabs.create({index: 1, url: url});
            });
        });
    } else {
        chrome.tabs.query( { index: 1 }, function (tabs) {
            if ( tabs[0] ) {
                if ( tabs[0].active ) {
                    chrome.tabs.update( tabs[0].id, { url: url });
                    return;
                 } else {
                     chrome.tabs.remove(tabs[0].id);
                 }
            }
            chrome.tabs.create({index: 1, url: url});
        });
    }

    if ( flags.firstLoadRetry )
        stopReloadOnError();
}

/*
 * Open Jmuconfig on the settings tab. Autologin if possible
 */
function openJMuconfig() {
   if ( unixUser && unixPass )
      openOnSecondTab('http://' + unixUser + ':' + unixPass + '@127.0.0.1:8080');
   else
      openOnSecondTab('http://127.0.0.1:8080');
}

/*
 * Open settings tab. Requires to login by default
 */
function openOptions() {
   if (flags.disableOptions)
      return;

   if (flags.disableLogin)
      openOnSecondTab("chrome://settings");
   else
      openOnSecondTab("options.html");

   errorCount = openOptionRetries +1;
   handleResumeReload();
}

function clearCacheAndReload() {
   chrome.browsingData.removeCache({}, function() {
      chrome.tabs.query( {}, function (tabs) {
         tabs.forEach(function(tab) {
            // Close settings tab
            if (tab.index == 1)
               chrome.tabs.remove(tab.id);
            else
               chrome.tabs.reload(tab.id);
         });
      });
   });
}
